/* Rotary Input library
 
Reinhard Kopka 29.12.2018

r_factor may be 1 or 2 depending on puls forming of device

*/
  
#ifndef DEF_INCLUDE_ROTARY_INPUT
#define DEF_INCLUDE_ROTARY_INPUT

// include the library code:
#include "PinChangeInterrupt.h"

// output - will be changed by interrupts - function for capsulation possible, but slower
volatile int16_t encoderPos = 0;    // this variable stores our current value of encoder position. 
                                    // Change to other type depending on wanted range
volatile int16_t encoderPosVal = 0; // this variable stores our current value of the corrected encoder position. 
                                    
// internal
static uint8_t last_state = 3;
static uint8_t last_cnt = 0;
static uint8_t new_state;
static int16_t r_min=0, r_max=255, r_start=0, r_step=1, r_factor=2, r_flag=0; // Configuration for counting type
static uint8_t pinA; // Our first hardware interrupt pin is digital pin 2
static uint8_t pinB; // Our second hardware interrupt pin is digital pin 3
static int16_t oldEncPos = -12345;  //stores the last encoder position value so we can compare to the current reading and see if it has changed (so we know when to print to the serial monitor)

// functions
void rotary_set  (int16_t min, int16_t max, int16_t start, int16_t step, int16_t factor, uint8_t flag); // configure counting
void rotary_reset(void);    // set count to configured start value
void rotary_reset(int val); // set count to val
void rotary_init (uint8_t pinA_, uint8_t pinB_);     // init interrupts
// internal interrupt
static void rotary_check(void);


// not all pins can be input pins, they need interrupt capability - see attachPinChangeInterrupt() docu
void rotary_init(uint8_t pinA_, uint8_t pinB_)
{
  pinA=pinA_;
  pinB=pinB_;
  pinMode(pinA, INPUT_PULLUP); // set pinA as an input, pulled HIGH to the logic voltage (5V or 3.3V for most cases)
  pinMode(pinB, INPUT_PULLUP); // set pinB as an input, pulled HIGH to the logic voltage (5V or 3.3V for most cases)
  
  attachPinChangeInterrupt(digitalPinToPinChangeInterrupt(pinA), rotary_check, CHANGE);
  attachPinChangeInterrupt(digitalPinToPinChangeInterrupt(pinB), rotary_check, CHANGE);
  last_state= 1*digitalRead(pinA) | 2*digitalRead(pinB);
  last_cnt=last_state;
}

// Interrupt function for the input pins
// Both Pin Interrupts call the same function on change, which just analyzes the pins
static void rotary_check()
{  
  cli(); //stop interrupts happening before we read pin values
  new_state= 1*digitalRead(pinA) | 2*digitalRead(pinB);

  if ( (new_state^last_cnt)==(1 | 2 ) )  // beide Bits müssen anders sein als beim letzten Zählen - warum ?
  {
    if ( (new_state ^ last_state) == 2)  // Änderung gegen letzten Status
      encoderPos+=r_step;
    else
      encoderPos-=r_step;
    last_cnt=new_state;

    encoderPosVal = encoderPos / r_factor;    
    if      (encoderPosVal>r_max) 
      if (r_flag & 1)  encoderPos = r_min*r_factor;
      else             encoderPos = r_max*r_factor;
    else if (encoderPosVal<r_min) 
      if (r_flag & 2)  encoderPos = r_max*r_factor;
      else             encoderPos = r_min*r_factor;
    encoderPosVal = encoderPos / r_factor;          
  }
  last_state=new_state;
  sei(); //restart interrupts
}

// nach dem Neustart geht es nach oben erst mit dem zweiten Dreh !!

// set the encoder values
// flag: 1-left circular 2-right circular
void rotary_set(int16_t min, int16_t max, int16_t start, int16_t step, int16_t factor, uint8_t flag)
{
  r_min=min;
  r_max=max;
  r_step=step;
  r_start=start; 
  r_factor = factor;
  r_flag = flag;
  encoderPosVal = r_start;
  encoderPos = encoderPosVal * r_factor;
}

// set encoder value back to start value
void rotary_reset(void)
{
  encoderPosVal = r_start;  
  encoderPos = encoderPosVal * r_factor;
}

// set encoder value back to start value
void rotary_reset(int val)
{
  encoderPosVal = val;  
  encoderPos = val * r_factor;
}

// true if encoder changed since last call, first call will always(?) be true
bool rotary_haschanged(void)
{
 // Wert wird außerhalb geändert, hier wird nur eine Änderung detektiert
  if (oldEncPos != encoderPosVal) 
  {
    oldEncPos = encoderPosVal;
    return true;
  }
  return false;
}

#endif  //DEF_INCLUDE_ROTARY_INPUT
